/*
 * Game Template
 * Copyright (C) 2007 Michael Birken
 * 
 * This file is part of Game Template.
 *
 * Game Template is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Game Template is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

namespace GameTemplate {
  public partial class GameTemplateForm : Form {

    [StructLayout(LayoutKind.Sequential)]
    public struct PeekMsg {
      public IntPtr hWnd;
      public Message msg;
      public IntPtr wParam;
      public IntPtr lParam;
      public uint time;
      public System.Drawing.Point p;
    }

    [System.Security.SuppressUnmanagedCodeSecurity]
    [DllImport("User32.dll", CharSet = CharSet.Auto)]
    public static extern bool PeekMessage(out PeekMsg msg, IntPtr hWnd,
             uint messageFilterMin, uint messageFilterMax, uint flags);

    const int WIDTH = 640;
    const int HEIGHT = 480;
    const int FRAMES_PER_SECOND = 60;

    const int BALLS = 5;
    const int BALL_SIZE = WIDTH / 10;
    const int PADDLE_WIDTH = BALL_SIZE * 2;
    const int PADDLE_HEIGHT = HEIGHT / 20;
    const int BALL_VELOCITY = 2;
    const int PADDLE_VELOCITY = 10;

    public long FREQUENCY = Stopwatch.Frequency;
    public long TICKS_PER_FRAME = Stopwatch.Frequency / FRAMES_PER_SECOND;
    private Bitmap bitmap;
    private Graphics g;
    public Graphics graphics;
    public Stopwatch stopwatch = new Stopwatch();
    public long nextFrameStart;
    public bool leftKeyPressed;
    public bool rightKeyPressed;
    public bool fullScreenMode;
    private Random random = new Random();
    private Ball[] balls = new Ball[BALLS];
    private int paddleX;

    public GameTemplateForm() {
      InitializeComponent();
      InitFrame();
    }

    private void GameTemplateForm_Load(object sender, EventArgs e) {
      InitModel();
      InitImage();
      stopwatch.Start();
      nextFrameStart = stopwatch.ElapsedTicks;
      Application.Idle += OnApplicationIdle;
    }

    private void OnApplicationIdle(object sender, EventArgs e) {
      PeekMsg msg;
      while (!PeekMessage(out msg, IntPtr.Zero, 0, 0, 0)) {
        do {
          UpdateModel();
          nextFrameStart += TICKS_PER_FRAME;
        } while (nextFrameStart < stopwatch.ElapsedTicks);
        RenderFrame();        
        long remainingTicks = nextFrameStart - stopwatch.ElapsedTicks;
        if (remainingTicks > 0) {
          Thread.Sleep((int)(1000 * remainingTicks / FREQUENCY));
        }
      }
    }

    private void UpdateModel() {
      // Move the balls  
      for (int i = 0; i < BALLS; i++) {
        Ball ball = balls[i];
        if (ball.x >= WIDTH - BALL_SIZE) {
          ball.vx = -BALL_VELOCITY;
        } else if (ball.x <= 0) {
          ball.vx = BALL_VELOCITY;
        }
        if (ball.y >= HEIGHT - BALL_SIZE - PADDLE_HEIGHT) {
          ball.vy = -BALL_VELOCITY;
        } else if (ball.y <= 0) {
          ball.vy = BALL_VELOCITY;
        }
        ball.x += ball.vx;
        ball.y += ball.vy;
      }

      // Move the paddle
      if (leftKeyPressed && paddleX > 0) {
        paddleX -= PADDLE_VELOCITY;
      }
      if (rightKeyPressed && paddleX < WIDTH - PADDLE_WIDTH) {
        paddleX += PADDLE_VELOCITY;
      }
    }

    private void RenderFrame() {
      g.FillRectangle(Brushes.Black, 0, 0, WIDTH, HEIGHT);

      // Render balls
      for (int i = 0; i < BALLS; i++) {
        Ball ball = balls[i];
        g.FillEllipse(Brushes.Red, ball.x, ball.y, BALL_SIZE, BALL_SIZE);
      }

      // Render paddle
      g.FillRectangle(Brushes.Yellow,
          paddleX, HEIGHT - PADDLE_HEIGHT, PADDLE_WIDTH, PADDLE_HEIGHT);

      // Draw to screen
      graphics.DrawImageUnscaled(bitmap, 0, 0);
    }

    private void InitFrame() {
      ClientSize = new Size(WIDTH, HEIGHT);
      StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
      SetStyle(ControlStyles.UserPaint, true);
      SetStyle(ControlStyles.AllPaintingInWmPaint, true);
    }

    private void InitModel() {
      for (int i = 0; i < BALLS; i++) {
        balls[i] = new Ball(
          random.Next(WIDTH - BALL_SIZE),
          random.Next(HEIGHT - BALL_SIZE - PADDLE_HEIGHT),
          random.Next(2) == 0 ? -BALL_VELOCITY : BALL_VELOCITY,
          random.Next(2) == 0 ? -BALL_VELOCITY : BALL_VELOCITY
        );
      }
      paddleX = (WIDTH - PADDLE_WIDTH) / 2;
    }

    private void InitImage() {
      graphics = CreateGraphics();
      bitmap = new Bitmap(WIDTH, HEIGHT, graphics);
      g = Graphics.FromImage(bitmap);
    }

    private void GameTemplateForm_KeyDown(object sender, KeyEventArgs e) {
      switch (e.KeyCode) {
        case Keys.Left:
          leftKeyPressed = true;
          break;
        case Keys.Right:
          rightKeyPressed = true;
          break;
      }
    }

    private void GameTemplateForm_KeyUp(object sender, KeyEventArgs e) {
      switch (e.KeyCode) {
        case Keys.Left:
          leftKeyPressed = false;
          break;
        case Keys.Right:
          rightKeyPressed = false;
          break;
        case Keys.F11:
          ToggleFullScreenMode();
          break;
      }
    }

    private void ToggleFullScreenMode() {
      fullScreenMode = !fullScreenMode;
      if (fullScreenMode) {
        ResolutionManager.EnterFullScreenMode(this, WIDTH, HEIGHT);
      } else {
        ResolutionManager.ExitFullScreenMode(this);
      }
      nextFrameStart = stopwatch.ElapsedTicks + TICKS_PER_FRAME;
    }
  }

  public class Ball {
    public int x;
    public int y;
    public int vx;
    public int vy;

    public Ball(int x, int y, int vx, int vy) {
      this.x = x;
      this.y = y;
      this.vx = vx;
      this.vy = vy;
    }
  }
}